# This file is copyright 2011, by Bill Cox, and released under the LGPL license,
# version 2, or (at your option) any later version.
# Modified in 2020 by Isaac Porat to comply with python 3 and NVDA version 2019.3.1

import os, socket, subprocess, time, threading
from queue import Queue

# These are used as a message level passed to the logFunc.
INFO_MESSAGE = 0
WARNING_MESSAGE = 1
IO_MESSAGE = 2
DEBUG_MESSAGE = 3

sharedSpeechHubInstance = None
speechHubRefCount = 0

class ResponseReader(threading.Thread):
    def __init__(self, sh, processingMessageLock, indexLock, queue):
        threading.Thread.__init__(self)
        self.sh = sh
        self.processingMessageLock = processingMessageLock
        self.indexLock = indexLock
        self.queue = queue
        self.timeToDie = False
        self.buffer = ""

    def run(self):
        while not self.timeToDie:
            data = self.recvResponse()
            self.queue.put(data)

    def quit(self):
        self.timeToDie = True

    def readline(self):
        """Read one whole line from the socket."""
        pointer = self.buffer.find("\n")
        while pointer == -1:
            try:
                d = (self.sh.socket.recv(1024)).decode("utf-8")
            except:
                #raise IOError
                raise
            if len(d) == 0:
                raise IOError
            self.buffer += d
            pointer = self.buffer.find("\n")
        line = self.buffer[:pointer].rstrip()
        self.sh.log(IO_MESSAGE, "GOT: " + line)
        self.buffer = self.buffer[pointer+1:]
        return line

    def recvMessage(self):
        """Read server response or a callback and return the triplet (code, msg, data)."""
        data = []
        c = None
        while True:
            line = self.readline()
            if line == None:
                return 0, "", []
            if len(line) < 4:
                self.sh.log(WARNING_MESSAGE, "Malformed data received from server!")
                return 0, "", []
            code, sep, text = line[:3], line[3], line[4:]
            if not (code.isalnum() and (c is None or code == c) and sep in ('-', ' ')):
                self.sh.log(WARNING_MESSAGE, "Malformed data received from server!")
                return 0, "", []
            if sep == ' ':
                return int(code), text, tuple(data)
            data.append(text)

    def recvResponse(self):
        """Read messages until we see a response, rather than an END
        message.  Update the indexList when we see END."""
        code, msg, data = self.recvMessage()
        while code == 702:
            self.indexLock.acquire()
            if data[0] in self.sh.indexList:
                index = self.sh.indexList.pop(data[0])
                if index != None:
                    self.sh.lastIndex = index
                    if self.sh.speakEndNotificationFunc != None:
                        self.sh.speakEndNotificationFunc(index)
                    self.sh.log(DEBUG_MESSAGE, "Setting last spoken index to: %s" % index)
            pendingSpeakMessages = len(self.sh.indexList)
            self.indexLock.release()
            code, msg, data = self.recvMessage()
        if code//100 != 2:
            self.sh.log(WARNING_MESSAGE, "Error in response: " + msg)
            return None
        return data

def getSpeechHub(logFunc=None):
    # Returns    a global instance.
    global sharedSpeechHubInstance
    global speechHubRefCount
    if sharedSpeechHubInstance != None:
        sharedSpeechHubInstance.log(DEBUG_MESSAGE, "Using existing SpeechHub connection")
        speechHubRefCount += 1
        return sharedSpeechHubInstance
    sharedSpeechHubInstance = SpeechHub(logFunc)
    speechHubRefCount = 1
    sharedSpeechHubInstance.startServer()
    return sharedSpeechHubInstance

class SpeechHub:

    def __init__(self, logFunc=None, speakEndNotificationFunc=None):
        """The log parameter is a function that will be called for all I/O with
           Speech-Hub.  It can be None if no logging is required.
           speakEndNotificationFunc is a function called whenever speech-hub
           completes speaking a message for which addIndex has been called.
           Note that both of these functions can be called by the recieve
           thread, not just the thread calling this object."""
        # Keep track of all the index marks still in progress
        self.logFunc = logFunc
        self.speakEndNotificationFunc = speakEndNotificationFunc
        self.indexList = dict()
        self.lastIndex = None
        self.currentIndex = None
        self.voice = None
        self.variant = None
        self.rate = None
        self.pitch = None
        self.volume = None
        self.paused = False
        self.language = None
        self.outputModule = None
        self.voice = None
        self.spelling = False
        self.punctLevel = None
        self.alreadyCancelled = False

    def log(self, level, text):
        if self.logFunc != None:
            self.logFunc(level, text)

    def writeline(self, text, start=True):
        """Write a line of text to SH.  Usually this is the entire comand.
           The exception is when we write data in a SPEAK command."""
        if start:
            # The lock will be released by the last call to recvResponse in the
            # main thread
            self.processingMessageLock.acquire()
        try:
            self.log(IO_MESSAGE, "SENT: " + text)
            self.socket.sendall((text + "\n").encode("utf-8"))
        except socket.error:
            raise Exception(self, "Speech Hub connection lost.")

    def recvResponse(self, done=True):
        """Read a response from the read thread. Usually this is the end of
           processing a message, so we unlock the processingMessageLock.  The
           exception is during a SPEAK message."""
        data = self.readQueue.get()
        if done:
            self.processingMessageLock.release()
        return data

    def spawnServer(self):
        """Attempts to spawn the speech-hub server."""
        self.log(INFO_MESSAGE, "Spawning SpeechHub")
        subprocess.Popen(["wscript", "invisible.vbs", "SpeechHub_SpeechServer.bat"],
            cwd=os.path.dirname(__file__), stdin=None, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

    def initServer(self):
        """Send speech-hub initialization messages."""
        self.startReadThread()
        self.writeline("SET self CLIENT_NAME unknown:nvda:espeak")
        self.recvResponse()
        self.writeline("SET self NOTIFICATION end on")
        self.recvResponse()
        self.writeline("set self CAP_LET_RECOGN none")
        self.recvResponse()

    def startReadThread(self):
        """Start the read thread and the write thread."""
        # We are forced to start a second thread to respond to SSIP
        # notifications quickly.  The queue is for received messages, not
        # notifications.  In theory, it should never have more than one message.
        self.readQueue = Queue()
        # The indexLock lock is to synchronize manipulation of indexList and
        # lastIndex.
        self.indexLock = threading.Lock()
        self.processingMessageLock = threading.Lock()
        self.readThread = ResponseReader(self, self.processingMessageLock, self.indexLock,
            self.readQueue)
        self.readThread.start()

    def startServer(self):
        """Start the server if it's not already running, and connect to it."""

        self.log(DEBUG_MESSAGE, "Trying to connect")
        socketConnectArgs = ("localhost", 6564)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # In NVDA 2012.2beta2, somehow they change the default to having a timeout
        self.socket.settimeout(None)
        try:
            self.socket.connect(socketConnectArgs)
            self.initServer()
            self.log(DEBUG_MESSAGE, "Connect succeeded")
        except socket.error as ex:
            self.spawnServer()
            # Give the server up to five seconds to come up
            totalDelay = 0.0
            delay = 0.25
            while totalDelay < 5.0:
                try:
                    self.socket.connect(socketConnectArgs)
                    self.initServer()
                    self.log(DEBUG_MESSAGE, "Connect succeeded")
                    return
                except socket.error as ex:
                    time.sleep(delay)
                    totalDelay = totalDelay + delay
            self.log(DEBUG_MESSAGE, "Connect failed\n")
            raise IOError

    def stopServer(self):
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        self.socket.close()

    def getLastIndex(self):
        # Check that we've read any pending speak response
        self.log(DEBUG_MESSAGE, "Client checked for last index mark (%s)" % self.lastIndex)
        return self.lastIndex

    def sendText(self, text):
        """Send multiline text but dont read the response."""
        text = text.replace("\r", "")
        # Some TTS engines don't strip control characters and this causes words
        # to be concatenated.
        text = text.replace("\n", " ")
        if text == ".":
            text = ".."
        text = text + "\n."
        try:
            self.writeline(text, False)
        except socket.error:
            raise Exception(self, "Speech-Hub connection lost.")

    def speak(self, text, charMode=False):
        """Speak the text."""
        if self.paused:
            # Need to unpause
            self.pause(False)
        # Now, aquire the processingMessageLock
        self.processingMessageLock.acquire()
        if not charMode:
            self.writeline("SPEAK", False) # We already aquired processingMessageLock
            self.recvResponse(False) # Keep the lock
            self.sendText(text)
        else:
            self.writeline("CHAR " + text, False)
        data = self.recvResponse(False) # For the response to the message text
        self.indexLock.acquire()
        if len(data) > 0:
            index = data[0]
            self.indexList[index] = self.currentIndex
        self.currentIndex = None
        self.indexLock.release()
        self.alreadyCancelled = False
        self.processingMessageLock.release()

    def sendChar(self, char):
        """Send a character to be spoken as a character."""
        if char == " ":
            self.speak("space", True)
        else:
            self.speak(char, True)

    def addIndex(self, index):
        self.currentIndex = index
        self.log(DEBUG_MESSAGE, "Client added index %s" % index)

    def setSpellMode(self, state):
        if self.spelling != state:
            if state:
                self.writeline("SET self SPELLING on")
            else:
                self.writeline("SET self SPELLING off")
            self.spelling = state
            self.recvResponse()

    def cancel(self):
        if self.alreadyCancelled:
            return
        self.writeline("CANCEL")
        self.recvResponse()
        self.indexLock.acquire()
        self.indexList.clear()
        self.indexLock.release()
        self.lastIndex = None
        self.log(DEBUG_MESSAGE, "Setting last spoken index to None")
        self.alreadyCancelled = True

    def pause(self, switch):
        if self.paused != switch:
            #if switch:
            #    self.writeline("PAUSE self")
            #else:
            #    self.writeline("RESUME self")
            self.paused = switch
            #self.recvResponse()

    def setRate(self, rate):
        if self.rate != rate:
            self.writeline("set self RATE %d" % rate)
            self.rate = rate
            self.recvResponse()

    def setPitch(self, pitch):
        if self.pitch != pitch:
            self.writeline("set self PITCH %d" % pitch)
            self.pitch = pitch
            self.recvResponse()

    def setVolume(self, volume):
        if self.volume != volume:
            self.writeline("set self VOLUME %d" % volume)
            self.volume = volume
            self.recvResponse()

    def listOutputModules(self):
        self.writeline("list OUTPUT_MODULES")
        return self.recvResponse()

    def setOutputModule(self, value):
        if self.outputModule != value:
            self.writeline("set self OUTPUT_MODULE " + value)
            response = self.recvResponse()
            if response == None:
                return False
            self.outputModule = value
            self.voice = None # We don't know what voice is now in effect
        return True

    def listVoices(self):
        self.writeline("list SYNTHESIS_VOICES")
        return self.recvResponse()

    def setVoice(self, voice):
        if self.voice != voice:
            self.writeline("set self SYNTHESIS_VOICE " + voice)
            self.voice = voice
            self.recvResponse()

    def setLanguage(self, language):
        language = language.replace('_','-')
        if self.language != language:
            self.writeline("set self LANGUAGE " + language)
            self.language = language
            self.recvResponse()

    def setPunctuation(self, level):
        """Set the punctuation level.  Valid valuds are 'none', 'some', and 'all'. """
        if self.punctLevel != level:
            self.writeline("set self PUNCTUATION " + level)
            self.punctLevel = level
            self.recvResponse()
        
    def quit(self):
        global sharedSpeechHubInstance
        global speechHubRefCount
        speechHubRefCount -= 1
        if speechHubRefCount > 0:
            return
        sharedSpeechHubInstance = None
        self.readThread.quit()
        self.writeline("quit")
        #self.recvResponse()
        self.readThread.join()
        self.stopServer()

    def setVariant(self, variant):
        if self.variant != variant:
            self.writeline("set self VOICE " + variant)
            self.variant = variant
            self.recvResponse()

    def listVariants(self):
        self.writeline("list VOICES")
        return self.recvResponse()

    def setRateBoost(self, value):
        """Set the Sonic rate boost mode on or off."""
        if value:
            self.writeline("set self RATE_BOOST_MODE ON")
        else:
            self.writeline("set self RATE_BOOST_MODE OFF")
        self.recvResponse()

    def getModuleProperty(self, moduleName, propertyName):
        self.writeline("get ENGINE_FACTORY_PROPERTY %s %s" % (moduleName, propertyName))
        return self.recvResponse()[0]

# This is appended to _speechhub.py to create the actual installTasks.py use
# in the plugin.  This is because I can't figure out how to load another module
# from installTasks.py

import logHandler

def log(level, text):
    """This function is passed to the speech-hub driver so it can interface to
       our logger."""
    if level == INFO_MESSAGE:
        logHandler.log.info(text)
    elif level == WARNING_MESSAGE:
        logHandler.log.debugWarning(text)
    elif level == IO_MESSAGE:
        logHandler.log.io(text)
    elif level == DEBUG_MESSAGE:
        logHandler.log.debug(text)
    else:
        logHandler.log.error("Invalid logging level:" + text)

log(DEBUG_MESSAGE, "Loading Speech-Hub post install script")

def readFile(fileName):
    """Read a file into an array of strings."""
    try:
        file = open(fileName, "r")
        lines = file.readlines()
        file.close()
    except IOError:
        raise Exception(None, "Unable to read " + fileName)
    log(DEBUG_MESSAGE, "Read " + fileName + " successfully")
    return lines

def writeFile(fileName, lines):
    """Write the lines to the file."""
    try:
        file = open(fileName, "w")
        file.writelines(lines)
        file.close()
    except IOError:
        raise Exception("Can\'t write to file " + fileName)
    log(DEBUG_MESSAGE, "Wrote " + fileName + " successfully")

def genScript(id, niceName, supportsVariants, supportsRateBoost):
    """This function generates the NVDA script to support a synth."""
    fileName = os.path.join(os.path.dirname(__file__), "sh_template")
    lines = readFile(fileName)
    newLines = []
    for line in lines:
        if supportsVariants or line.find('SynthDriver.VariantSetting') == -1:
            if supportsRateBoost or line.find('BooleanDriverSetting("rateBoost"') == -1:
                newLine = line.replace('MODULE_NAME', id)
                newLine = newLine.replace('SHOWN_NAME', niceName)
                newLines.append(newLine)
    outputFileName = "sh_" + id + ".py"
    outputFileName = os.path.join(os.path.dirname(__file__), "synthDrivers", outputFileName)
    log(IO_MESSAGE, "Generating " + outputFileName)
    writeFile(outputFileName, newLines)

def getSynthSettings():
    log(WARNING_MESSAGE, "Checked that speechhub MODULE_NAME is up")
    sh = getSpeechHub(log)
    sh.startServer()
    modules = sh.listOutputModules()
    synths = {}
    for module in modules:
        caption = sh.getModuleProperty(module, "engine.caption")
        if caption.upper() == "NONE":
            log(DEBUG_MESSAGE, "No support for module " + module)
        else:
            audioData = sh.getModuleProperty(module, "engine.supports.audio.data")
            useSonic = sh.getModuleProperty(module, "engine.use.sonic")
            supportsVariants = sh.getModuleProperty(module, "engine.supports.variants")
            synths[module] = [caption, supportsVariants.lower() == "true",
                audioData.lower() == "true" and useSonic.lower() == "false"]
    sh.quit()
    sh.stopServer()
    return synths

def onInstall():
    log(INFO_MESSAGE, "Installing SpeechHub add-on")
    synths = getSynthSettings()
    for id in synths:
        niceName, supportsVariants, supportsRateBoost = synths[id]
        log(DEBUG_MESSAGE, "Name " + niceName + " Variants " + str(supportsVariants) +
            " rate boost " + str(supportsRateBoost))
        genScript(id, niceName, supportsVariants, supportsRateBoost)
